"""
Across import machinery, full of deadly traps and zombie unicorns. Stay away.
"""

import sys
import importlib
import importlib.util
import types
import ast
import pickle
import functools
import logging
import socket
import time


# Basic logging utilities. Since importer needs to be self-contained, they must be implemented here.

_debug_level = 0
_debug_handler = None

logger = logging.getLogger('across')
if not logger.handlers:
    logger.addHandler(logging.NullHandler())


def get_debug_level():
    return _debug_level


def set_debug_level(level):
    global _debug_level
    global _debug_handler

    if level <= 0:
        if _debug_handler is not None:
            logger.removeHandler(_debug_handler)
            logger.setLevel(logging.NOTSET)
            _debug_handler = None
    else:
        if _debug_handler is None:
            formatter = logging.Formatter(
                '%(asctime)s %(levelname)-5s ' + socket.gethostname().replace('%', '%%') +
                '/%(process)d %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S',
            )
            formatter.converter = time.gmtime

            _debug_handler = logging.StreamHandler()
            _debug_handler.setFormatter(formatter)
            logger.addHandler(_debug_handler)

        if level == 1:
            logging_level = logging.INFO
        else:
            logging_level = logging.DEBUG

        logger.setLevel(logging_level)
        _debug_handler.setLevel(logging_level)

    _debug_level = level


# Compile given source, but skip the code inside "if __name__ == '__main__'". When such "if" statement cannot
# be located, return None to indicate that the source cannot be executed without triggerring the usual effects
# on main module
def _compile_safe_main(source, filename):
    mod = compile(source, filename, 'exec', ast.PyCF_ONLY_AST, dont_inherit=True)
    if mod.body and isinstance(mod.body[-1], ast.If):
        test = mod.body[-1].test
        if (
            isinstance(test, ast.Compare) and
            isinstance(test.left, ast.Name) and
            test.left.id == '__name__' and
            len(test.ops) == 1 and
            isinstance(test.ops[0], ast.Eq) and
            isinstance(test.comparators[0], ast.Str) and
            test.comparators[0].s == '__main__'
        ):
            mod.body[-1].test = ast.Num(0, lineno=0, col_offset=0)
            return compile(mod, filename, 'exec', dont_inherit=True)

    return None


def _find_across_loader(fullname):
    if fullname == '__main__':
        # Let's deal with various quirks of runpy module:
        #   - __spec__ is None when using 'python -c script.py', we have to fall back to __loader__
        #   - __spec__.name differs from fullname when using 'python -m modname'; __spec__.name is the one
        #     expected by __spec__.__loader__
        main_mod = sys.modules.get('__main__')
        if main_mod is not None:
            spec = getattr(main_mod, '__spec__', None)
            if spec is None:
                loader = getattr(main_mod, '__loader__', None)
                if loader is None:
                    raise ValueError('__main__.__loader__ is not set')
                logger.debug('Querying module %r, loader=%r', fullname, loader)
                return _interrogate_loader(loader, fullname)
            else:
                if spec.loader is None:
                    raise ValueError('{!r}.loader is not set'.format(spec))
                logger.debug('Querying module %r (main), spec=%r', spec.name, spec)
                return _interrogate_loader(spec.loader, spec.name)

    try:
        spec = importlib.util.find_spec(fullname)
    except ImportError:
        logger.debug('Module %r not found', fullname)
        return None
    if spec is None:
        logger.debug('Module %r not found', fullname)
        return None
    if spec.loader is None:
        # Yet another hack, this time for implicit namespace packages.
        if spec.submodule_search_locations:
            logger.debug('Querying module %r (namespace), spec=%r', fullname, spec)
            return AcrossLoader('', True, '<source>')
        raise ValueError('{!r}.loader is not set'.format(spec))
    logger.debug('Querying module %r, spec=%r', fullname, spec)
    return _interrogate_loader(spec.loader, fullname)


def _interrogate_loader(loader, fullname):
    if isinstance(loader, AcrossLoader):
        return loader
    get_source = getattr(loader, 'get_source', None)
    if get_source is None:
        raise TypeError('Loader {!r} for module {} does not implement get_source method'.format(loader, fullname))
    source = get_source(fullname)
    if source is None:
        raise ValueError('Source is not available for loader {!r} and module {}'.format(loader, fullname))
    is_package = getattr(loader, 'is_package', None)
    if is_package is None:
        raise TypeError('Loader {!r} for module {} does not implement is_package method'.format(loader, fullname))
    package = is_package(fullname)
    get_filename = getattr(loader, 'get_filename', None)
    if get_filename is None:
        raise TypeError('Loader {!r} for module {} does not implement get_filename method'.format(loader, fullname))
    filename = get_filename(fullname)
    if filename is None:
        raise ValueError('Filename is not available for loader {!r} and module {}'.format(loader, fullname))
    return AcrossLoader(source, package, filename)


# For packages, __path__ attribute holds a list of directories where submodules are located. This object
# acts as a replacement of that list. First, it allows detecting if parent module was loaded by us. Second,
# it prevents mutating __path__ by user code.
class AcrossSearchPath:
    def __getitem__(self, item):
        raise IndexError

    def __iter__(self):
        return iter(())

    def __len__(self):
        return 0

    def __contains__(self, item):
        return False


class AcrossFinder:
    def __init__(self):
        self.__loaders = {}
        self.__conn = None
        self.__exported_modules = set()

    def export(self, modules, loaders=None):
        for name in modules:
            if '.' in name:
                raise ValueError('Not a top-level module: {}'.format(name))
            if name in sys.modules and name not in self.__exported_modules and name != '__main__':
                raise ValueError('Cannot export module {} because it is already imported'.format(name))

        had_exported_modules = bool(self.__exported_modules)
        had_main = ('__main__' in self.__exported_modules)
        self.__exported_modules.update(modules)
        if loaders is not None:
            self.__loaders.update(loaders)

        if self.__exported_modules and not had_exported_modules:
            sys.meta_path.insert(0, self)
        if '__main__' in self.__exported_modules and not had_main:
            # I originally injected fake module here with __getattr__ loading the module lazily, but that
            # didn't work correctly on CPython 3.7 (ImportError raised when _compile_safe_main returns None
            # was swallowed by interpreter).
            sys.modules.pop('__main__', None)

    def set_connection(self, conn):
        if self.__conn is not None:
            raise ValueError('Cannot import modules over multiple connections')
        self.__conn = conn

    def find_spec(self, fullname, path=None, target=None):
        loader = self.find_module(fullname, path)
        if loader is None:
            return None
        spec = importlib.util.spec_from_loader(fullname, loader)
        if spec.submodule_search_locations is not None:  # is it a package?
            spec.submodule_search_locations = AcrossSearchPath()
        return spec

    def find_module(self, fullname, path=None):
        top_name, dot, _ = fullname.partition('.')
        # Skip if topmost module is not exported
        if top_name not in self.__exported_modules:
            return None
        # Skip if parent module was not loaded by us
        if dot and not isinstance(path, AcrossSearchPath):
            return None

        if fullname not in self.__loaders:
            if self.__conn is None:
                return None
            self.__loaders[fullname] = self.__conn.call(_find_across_loader, fullname)
        return self.__loaders[fullname]


class AcrossLoader:
    def __init__(self, source, package, filename, code=None):
        self.__source = source
        self.__package = package
        self.__filename = filename
        self.__code = code

    def get_filename(self, fullname):
        if self.__filename.startswith('<') and self.__filename.endswith('>'):
            return self.__filename
        # We need to apply some sort of mangling to prevent clash with local files. Appending '*' seems like
        # a good solution, because it's rather unlikely that anyone else would create a file with '.py*' extension.
        return self.__filename + '*'

    def get_code(self, fullname):
        if self.__code is None:
            if fullname == '__main__':
                code = _compile_safe_main(self.get_source(fullname), self.get_filename(fullname))
                if code is None:
                    raise ImportError('__main__ module cannot be safely imported remotely because it lacks '
                                      'proper "if __name__ == \'__main__\'" guard')
            else:
                code = compile(self.get_source(fullname), self.get_filename(fullname), 'exec', dont_inherit=True)
            self.__code = code
        return self.__code

    def get_source(self, fullname):
        return self.__source

    def is_package(self, fullname):
        return self.__package

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        exec(self.get_code(module.__name__), module.__dict__)

    def __reduce__(self):
        return AcrossLoader, self.deconstruct(with_code=False)

    # For internal use only
    def deconstruct(self, with_code):
        return self.__source, self.__package, self.__filename, (self.__code if with_code else None)


# This list contains 'across' module with all its depenendencies, excluding importer module
# (which is a different story, because it needs to load itself).
_core_module_names = (
    'across',
    'across._utils',
    'across._channels',
)


def get_bootloader(func, *args):
    own_loader = _find_across_loader(__name__)
    if own_loader is None:
        raise ImportError('{} not found'.format(__name__))

    loaders = {}
    for module_name in _core_module_names:
        loader = _find_across_loader(module_name)
        if loader is None:
            raise ImportError('{} not found'.format(module_name))
        loaders[module_name] = loader.deconstruct(with_code=False)

    data = {
        'own_loader': own_loader.deconstruct(with_code=False),
        'ns': {'__name__': __name__},
        'loaders': loaders,
        'debug_level': get_debug_level(),
        'func_with_args': pickle.dumps((func, args), protocol=3),
    }

    return ("__across_boot={!r};exec(__across_boot['own_loader'][0],__across_boot['ns']);"
            "__across_boot=__across_boot['ns']['_bootstrap'](__across_boot);__across_boot(ACROSS)\n".format(data))


def _module_from_spec(spec):
    module = types.ModuleType(spec.name)
    module.__spec__ = spec
    module.__file__ = spec.origin
    module.__loader__ = spec.loader
    module.__package__ = spec.parent
    if spec.submodule_search_locations is not None:
        module.__path__ = spec.submodule_search_locations
    return module


_finder = AcrossFinder()


def get_finder():
    return _finder


def _bootstrap(data):
    # Enable debug mode inside bootloader as soon as possible.
    debug_level = data['debug_level']
    set_debug_level(debug_level)

    logger.debug('Bootloader starts, python=%r', sys.version)

    # The environment in which we are currently running code is not a proper Python module: __file__/__package__
    # are not set, there is no entry in sys.modules, etc. First step is to create our own module.
    if __name__ in sys.modules:
        raise RuntimeError('{} already in sys.modules'.format(__name__))
    # Beware: AcrossLoader.__module__ is invalid. We will have to recreate loader object later.
    tmp_loader = AcrossLoader(*data['own_loader'])
    own_spec = importlib.util.spec_from_loader(__name__, tmp_loader)
    own_module = sys.modules[__name__] = _module_from_spec(own_spec)
    tmp_loader.exec_module(own_module)

    # Our own module is ready now. We should stop using current globals and switch to the ones from 'own_module'.
    # Start with switching debug mode.
    set_debug_level(0)
    own_module.set_debug_level(debug_level)

    # Recreate our own loader.
    own_loader = own_module.AcrossLoader(*tmp_loader.deconstruct(with_code=True))
    own_module.__loader__ = own_spec.loader = own_loader

    # The most tricky part is behind us. Now let's set up finder object for loading remaining modules.
    loaders = dict(
        (module_name, own_module.AcrossLoader(*args))
        for module_name, args in data['loaders'].items()
    )
    loaders[__name__] = own_loader
    own_module.get_finder().export(['across'], loaders=loaders)

    # Now we can finally load startup function that takes care of creating Connection object.
    func, args = pickle.loads(data['func_with_args'])
    func_with_args = functools.partial(func, *args)

    logger.debug('Bootloader ends')

    return func_with_args



import argparse

from . import __version__, Connection
from .servers import run_tcp, run_unix, BIOSConnectionHandler
from ._utils import set_debug_level


def _parse_args():
    parser = argparse.ArgumentParser(
        description='Establish across connections.',
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__,
    )
    parser.add_argument(
        '--server',
        action='store_true',
        help='switch to server mode: listen for incoming connections',
    )
    parser.add_argument(
        '--tcp',
        metavar='HOST:PORT',
        help='connect over TCP',
        type=_parse_tcp,
        dest='address',
    )
    parser.add_argument(
        '--unix',
        metavar='PATH',
        help='connect over Unix domain socket',
        type=lambda arg: ('unix', arg),
        dest='address',
    )
    parser.add_argument(
        '--stdio',
        help='connect over stdin/stdout pipes',
        action='store_const',
        const=('stdio',),
        dest='address',
    )
    parser.add_argument(
        '--execute',
        metavar='COMMAND',
        help='execute Python code remotely',
        type=lambda arg: ('execute', arg),
        dest='action',
    )
    parser.add_argument(
        '--wait',
        help='block until connection is remotely closed',
        action='store_const',
        const=('wait',),
        dest='action'
    )
    parser.add_argument(
        '-v',
        help='print debug information to stderr; specify multiple times to increase verbosity',
        action='count',
        default=0,
        dest='verbose',
    )

    args = parser.parse_args()
    if not args.address:
        parser.error('address has not been specified')
    if args.server and args.action:
        parser.error('actions are not supported in server mode')
    if not args.server and not args.action:
        parser.error('action has not been specified')
    if args.server and args.address[0] == 'stdio':
        parser.error('cannot start server over stdin/stdout pipes')

    return args


def _parse_tcp(arg):
    try:
        host, sep, port = arg.rpartition(':')
        if ':' in host:  # IPv6 address, needs to be enclosed in []
            if not host.startswith('[') or not host.endswith(']'):
                raise ValueError
            host = host[1: -1]
        port = int(port)
        if '[' in host or ']' in host or not sep or not 0 <= port <= 65535:
            raise ValueError
        return 'tcp', host, port
    except ValueError:
        raise argparse.ArgumentTypeError('invalid value: {!r}'.format(arg)) from None


def _handle_args(args):
    set_debug_level(args.verbose)
    if args.server:
        _run_server(args)
    else:
        _run_client(args)


def _run_server(args):
    handler = BIOSConnectionHandler()
    if args.address[0] == 'tcp':
        run_tcp(args.address[1], args.address[2], handler=handler)
    elif args.address[0] == 'unix':
        run_unix(args.address[1], handler=handler)
    else:
        raise AssertionError(args.address)


def _run_client(args):
    if args.address[0] == 'tcp':
        conn = Connection.from_tcp(args.address[1], args.address[2])
    elif args.address[0] == 'unix':
        conn = Connection.from_unix(args.address[1])
    elif args.address[0] == 'stdio':
        conn = Connection.from_stdio()
    else:
        raise AssertionError(args.address)
    with conn:
        if args.action[0] == 'wait':
            conn.wait()
        elif args.action[0] == 'execute':
            obj = conn.execute(args.action[1])
            if obj is not None:
                print(obj)
        else:
            raise AssertionError(args.action)


def main():
    _handle_args(_parse_args())


if __name__ == '__main__':
    main()



"""
Various helpers. For internal use only.
"""

import collections
import traceback
import sys
import queue
import logging
import threading

from ._importer import logger, get_debug_level, set_debug_level


# Export utilities implemented in importer module
logger = logger
get_debug_level = get_debug_level
set_debug_level = set_debug_level


class _AtomicCounter:
    def __init__(self, start=0):
        self._value = start
        self._lock = threading.Lock()

    def __next__(self):
        with self._lock:
            value = self._value
            self._value += 1
            return value

    def __iter__(self):
        return self


atomic_count = _AtomicCounter


class IdentityAdapter(logging.LoggerAdapter):
    def __init__(self, logger, identity):
        super().__init__(logger, {})
        self._prefix = '[{}] '.format(identity)

    def process(self, msg, kwargs):
        return self._prefix + msg, kwargs


# Lightweight version of concurrent.futures.Future
class Future:
    def __init__(self):
        self._done = threading.Event()
        self._result = None
        self._exc = None

    def result(self):
        self._done.wait()
        if self._exc is not None:
            raise self._exc
        return self._result

    def set_result(self, result):
        self._result = result
        self._done.set()

    def set_exception(self, exc):
        self._exc = exc
        self._done.set()


# Runs functions in worker threads.
class Executor:
    def __init__(self, idle_timeout=10.0):
        self._lock = threading.Lock()
        self._tasks = collections.deque()
        self._idle_condition = threading.Condition(self._lock)
        self._num_idle_threads = 0
        self._num_threads = 0
        self._idle_timeout = idle_timeout
        self._joinable_thread = None
        self._closing = False
        self._close_condition = threading.Condition(self._lock)

    def submit(self, func, *args):
        with self._lock:
            if self._closing:
                raise ValueError('Executor is closed')
            if self._num_idle_threads > 0:
                self._tasks.append((func, args))
                self._num_idle_threads -= 1
                self._idle_condition.notify()
            else:
                thread = threading.Thread(target=self._thread_func, args=(func, args), daemon=True)
                thread.start()
                self._num_threads += 1

    def _thread_func(self, func, args):
        while True:
            try:
                func(*args)
            # All exceptions from threading.Thread.run() method are ignored, here we emulate that behavior.
            except BaseException:
                ignore_exception_at(func)

            with self._lock:
                if self._closing:
                    break
                self._num_idle_threads += 1
                while not self._tasks and not self._closing:
                    if not self._idle_condition.wait(self._idle_timeout) and self._num_idle_threads > 0:
                        self._num_idle_threads -= 1
                        break
                if not self._tasks:
                    break
                func, args = self._tasks.popleft()

        with self._lock:
            self._num_threads -= 1
            if self._closing and self._num_threads == 0:
                self._close_condition.notify_all()
            if self._joinable_thread:
                self._joinable_thread.join()
            self._joinable_thread = threading.current_thread()

    def close(self):
        with self._lock:
            self._closing = True
            self._num_idle_threads = 0
            self._idle_condition.notify_all()
            while self._num_threads > 0:
                self._close_condition.wait()
            if self._joinable_thread:
                self._joinable_thread.join()
                self._joinable_thread = None


# based on PyErr_WriteUnraisable
def ignore_exception_at(obj):
    try:
        print('Exception ignored in: {!r}'.format(obj), file=sys.stderr)
        traceback.print_exc()
    except OSError:
        pass


# Variant of traceback.format_exception_only that really formats only the exception. For SyntaxError,
# traceback.format_exception_only outputs also some additional lines that pretend to be a part of the traceback.
def format_exception_only(error):
    if isinstance(error, SyntaxError):
        return 'SyntaxError: {}'.format(error)
    return ''.join(traceback.format_exception_only(type(error), error)).strip()


# Synchronized queue implementation that can be safely used from within __del__ methods, as opposed to
# queue.Queue (https://bugs.python.org/issue14976). In Python 3.7, queue.SimpleQueue can be used instead.
#
# This class was written with CPython in mind, and may not work correctly with other Python implementations.
class SimpleQueue:
    def __init__(self):
        self.__lock = threading.Lock()
        self.__waiter = threading.Lock()
        self.__items = collections.deque()

        # Create bound method objects for later use in get()/put() - creating such objects may trigger GC,
        # which must not happen inside get()/put(). These methods are all written in C, and are expected not to
        # allocate GC memory internally (e.g. with PyObject_GC_New).
        self.__lock_acquire = self.__lock.acquire
        self.__lock_release = self.__lock.release
        self.__waiter_acquire = self.__waiter.acquire
        self.__waiter_release = self.__waiter.release
        self.__items_popleft = self.__items.popleft
        self.__items_append = self.__items.append

        self.__waiter_acquire()

    # Wait until a queue becomes non-empty, and retrieve a single element from the queue. This function may be
    # called only from a single thread at a time.
    def get(self, timeout=None):
        if timeout is None:
            self.__waiter_acquire()
        elif not self.__waiter_acquire(timeout=timeout):
            raise queue.Empty
        self.__lock_acquire()
        item = self.__items_popleft()
        if self.__items:
            self.__waiter_release()
        self.__lock_release()
        return item

    # Insert a single element at the end of the queue. This method may be safely called from multiple threads,
    # and from __del__ methods.
    def put(self, item):
        self.__lock_acquire()
        if not self.__items:
            self.__waiter_release()
        self.__items_append(item)
        self.__lock_release()



import threading
import socket
import subprocess
import sys
import pickle
import base64
import os.path
import signal
import time

from ._utils import ignore_exception_at, logger as _logger, get_debug_level, set_debug_level
from . import _get_bios_superblock, Connection, _sanitize_options


_windows = (sys.platform == 'win32')


class ConnectionHandler:
    def handle_socket(self, sock):
        raise NotImplementedError

    def cancel(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class LocalConnectionHandler(ConnectionHandler):
    def __init__(self, *, options=None):
        self.__lock = threading.Lock()
        self.__unclosed_conns = set()
        self.__stopped_conn = None
        self.__closed = False
        self.__cancelled = False
        self.__options = _sanitize_options(options)
        if self.__options.accept_exported_modules:
            raise ValueError('Cannot import modules over multiple connections')

    def handle_socket(self, sock):
        with self.__lock:
            if self.__closed:
                raise ValueError('Connection handler is closed')
            if self.__cancelled:
                raise ValueError('Connection handler is cancelled')
            conn = Connection.from_socket(sock, options=self.__options, on_stopped=self.__connection_stopped)
            self.__unclosed_conns.add(conn)

    def __close_stopped_connection_locked(self):
        self.__unclosed_conns.remove(self.__stopped_conn)
        self.__close_conn(self.__stopped_conn)
        self.__stopped_conn = None

    def __connection_stopped(self, conn):
        with self.__lock:
            if self.__closed:
                return
            if self.__stopped_conn:
                self.__close_stopped_connection_locked()
            self.__stopped_conn = conn

    def __close_conn(self, conn):
        try:
            conn.close()
        # Connection.close() may raise all different exception types.
        except Exception:
            ignore_exception_at(conn.close)

    def cancel(self):
        with self.__lock:
            self.__cancelled = True
            for conn in self.__unclosed_conns:
                conn.cancel()

    def close(self):
        with self.__lock:
            self.__closed = True
            if self.__stopped_conn:
                self.__close_stopped_connection_locked()
            conns = self.__unclosed_conns
            self.__unclosed_conns = []
        for conn in conns:
            self.__close_conn(conn)


_serve_arg = '_serve'


def _get_process_close_timeout(options):  # patched by tests
    return options.timeout


class _ProcessConnectionHandlerBase(ConnectionHandler):
    def __init__(self, *, options=None):
        self._procs = set()
        self._cancelled = False
        self._closed = False
        self._options = _sanitize_options(options)

    def handle_socket(self, sock):
        if self._closed:
            raise ValueError('Connection handler is closed')
        if self._cancelled:
            raise ValueError('Connection handler is cancelled')
        proc = self._create_proc(sock)
        _logger.debug('Started child process, pid=%r', proc.pid)
        self._procs.add(proc)
        self._collect_procs()

    if _windows:
        def _create_proc(self, sock):
            proc = subprocess.Popen(
                self._get_base_args(),
                stdin=subprocess.PIPE,
                stdout=subprocess.DEVNULL,
            )
            proc.stdin.write(sock.share(proc.pid))
            proc.stdin.close()
            sock.close()
            return proc

        # There's CTRL_C_EVENT on Windows, but it's restricted to processes running in a console window.
        # Since there's no better alternative, let's just kill the process.
        _interrupt_signal = signal.SIGTERM
    else:
        def _create_proc(self, sock):
            proc = subprocess.Popen(
                self._get_base_args() + [str(int(sock.family))],
                stdin=sock.fileno(),
                stdout=subprocess.DEVNULL,
            )
            sock.close()
            return proc

        _interrupt_signal = signal.SIGINT

    def _collect_procs(self):
        for proc in self._procs.copy():
            if proc.poll() is not None:
                _logger.debug('Joined process %r', proc.pid)
                self._procs.remove(proc)

    def _get_base_args(self):
        raise NotImplementedError

    def _interrupt_proc(self, proc):
        if self._cancelled:
            return
        if proc.poll() is None:
            _logger.debug('Interrupting process %r', proc.pid)
            try:
                proc.send_signal(self._interrupt_signal)
            except OSError:  # process already exited
                pass

    def _kill_proc(self, proc):
        if self._cancelled:
            return
        if proc.poll() is None:
            _logger.debug('Killing process %r', proc.pid)
            try:
                proc.kill()
            except OSError:  # process already exited
                pass

    def _wait_proc(self, proc, timeout):
        if timeout <= 0.0:
            return
        try:
            proc.wait(timeout)
        except subprocess.TimeoutExpired:
            pass

    def cancel(self):
        for proc in self._procs:
            self._kill_proc(proc)
        self._cancelled = True

    def close(self):
        self._closed = True
        for proc in self._procs:
            self._interrupt_proc(proc)
        deadline = time.time() + _get_process_close_timeout(self._options)
        for proc in self._procs:
            _logger.debug('Joining process %r', proc.pid)
            self._wait_proc(proc, deadline - time.time())
            self._kill_proc(proc)
            proc.wait()
            _logger.debug('Process joined')


class ProcessConnectionHandler(_ProcessConnectionHandlerBase):
    def _get_base_args(self):
        data = base64.b64encode(pickle.dumps((get_debug_level(), self._options))).decode('ascii')
        return [sys.executable, '-m', __name__, _serve_arg, data]


if _windows:
    _socket_bios = r"""import socket,sys
sock = socket.fromshare(sys.stdin.buffer.read())
"""
else:
    _socket_bios = r"""import os,sys,socket
sock = socket.fromfd(0, int(sys.argv[2]), socket.SOCK_STREAM)
devnull_fd = os.open(os.devnull, os.O_RDONLY)
os.dup2(devnull_fd, 0)
os.close(devnull_fd)
"""

_socket_bios += r"""import struct
sock.settimeout(float(sys.argv[1]))
sock.sendall({superblock!r})
def recvall(size):
    buf = b''
    while len(buf) < size:
        buf += sock.recv(size - len(buf)) or sys.exit(1)
    return buf
ACROSS = 'socket', sock
try:
    exec(recvall(struct.unpack('>I', recvall(20)[-4:])[0]))
except KeyboardInterrupt:
    raise SystemExit
""".format(superblock=_get_bios_superblock())


class BIOSConnectionHandler(_ProcessConnectionHandlerBase):
    def _get_base_args(self):
        return [sys.executable, '-c', _socket_bios, str(self._options.timeout)]


# IPPROTO_IPV6 is missing on Windows and Python < 3.8
try:
    _IPPROTO_IPV6 = socket.IPPROTO_IPV6
except AttributeError:
    if _windows:
        # https://github.com/mirror/mingw-w64/blob/16151c441e89081fd398270bb888511ebef6fb35/
        # mingw-w64-headers/include/winsock2.h
        _IPPROTO_IPV6 = 41
    else:
        raise


def _tune_server_socket(sock):
    if sock.family in (socket.AF_INET, socket.AF_INET6):
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if sock.family == socket.AF_INET6:
        sock.setsockopt(_IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)


_open_socket = socket.socket  # patched by tests


def _run_server(family, address, handler):
    if handler is None:
        handler = LocalConnectionHandler()
    sock = _open_socket(family)
    _tune_server_socket(sock)
    sock.bind(address)
    _logger.info('Listening on %r', address)
    sock.listen(socket.SOMAXCONN)
    try:
        while True:
            client_sock, client_addr = sock.accept()
            _logger.info('Accepted connection from %r', client_addr)
            handler.handle_socket(client_sock)
    except KeyboardInterrupt:
        _logger.info('Received interrupt')
    sock.close()
    handler.close()


def run_tcp(host, port, *, handler=None):
    addrinfo = socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
    if not addrinfo:
        # C getaddrinfo() returns one or more items, so this should hopefully never happen
        raise RuntimeError('getaddrinfo returned empty list')
    family, _, _, _, address = addrinfo[0]
    _run_server(family, address, handler)


def run_unix(path, *, handler=None):
    if not hasattr(socket, 'AF_UNIX'):
        raise RuntimeError('Unix domain sockets are not available')
    if os.path.exists(path):
        os.unlink(path)
    _run_server(socket.AF_UNIX, path, handler)
    os.unlink(path)


def main():
    # safeguard in case someone accidentally runs this module
    if len(sys.argv) < 2 or sys.argv[1] != _serve_arg:
        return

    debug_level, options = pickle.loads(base64.b64decode(sys.argv[2]))
    set_debug_level(debug_level)
    if _windows:
        sock = socket.fromshare(sys.stdin.buffer.read())
    else:
        family = int(sys.argv[3])
        sock = socket.fromfd(0, family, socket.SOCK_STREAM)
        devnull_fd = os.open(os.devnull, os.O_RDONLY)
        os.dup2(devnull_fd, 0)
        os.close(devnull_fd)
    try:
        with Connection.from_socket(sock, options=options) as conn:
            conn.wait()
    except KeyboardInterrupt:
        raise SystemExit


if __name__ == '__main__':
    main()



import logging

from . import get_connection, DisconnectError
from ._utils import logger as _across_logger


_across_logger_name = _across_logger.name
_default_formatter = logging.Formatter()


class AcrossHandler(logging.Handler):
    def __init__(self, conn=None):
        super(AcrossHandler, self).__init__()
        if conn is None:
            conn = get_connection()
        self.__conn = conn

    def emit(self, record):
        try:
            if record.name == _across_logger_name:
                return

            # there are a few things that need to be done before serializing record:
            #   - args may not be pickleable, so we need to format message locally;
            #   - tracebacks are not pickleable, so we need to format exception locally;
            #   - standard formatter defines some temporary attributes on records: record.message and
            #     record.asctime; as an optimization, we omit them during serialization
            if record.exc_info and not record.exc_text:
                record.exc_text = (self.formatter or _default_formatter).formatException(record.exc_info)
            attrs = record.__dict__.copy()
            message = attrs.pop('message', None)
            if message is None:
                message = record.getMessage()
            attrs['msg'] = message
            attrs['args'] = ()
            attrs['exc_info'] = None
            attrs.pop('asctime', None)

            self.__conn.call(_handle, attrs)
        except DisconnectError:
            pass
        except Exception:
            self.handleError(record)


def _handle(attrs):
    record = logging.makeLogRecord(attrs)
    logging.getLogger(record.name).handle(record)



from setuptools import setup

import across


setup(
    name='across',
    version=across.__version__,
    author='Krzysztof Rusek',
    author_email='savix5@gmail.com',
    description='Run Python code across different machines',
    license='MIT',
    packages=['across'],
    platforms=['any'],
    python_requires='>=3.4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import threading
import subprocess
import select
import os
import sys
import socket
import errno
import signal

from ._utils import logger as _logger


_windows = (sys.platform == 'win32')


class Channel:
    def set_timeout(self, timeout):
        pass

    def connect(self):
        pass

    def recv_into(self, buffer):
        raise NotImplementedError

    def send(self, buffer):
        raise NotImplementedError

    def cancel(self):
        pass

    def close(self):
        pass


def _make_timeout_error():
    return OSError(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT))


def _make_cancelled_error():
    return OSError(errno.ECANCELED, os.strerror(errno.ECANCELED))


if _windows:
    class _Poller:
        def __init__(self):
            self.__pipe = None
            self.__cancel_lock = threading.Lock()
            self.__cancelled = False
            self.__send_fd = None
            self.__recv_fd = None
            self.__timeout = None

        def __create_pipe(self):
            sock = socket.socket()
            sock.bind(('127.0.0.1', 0))
            sock.listen(1)
            in_pipe = socket.socket()
            in_pipe.connect(sock.getsockname())
            out_pipe = sock.accept()[0]
            sock.close()
            self.__pipe = in_pipe, out_pipe

        def set_timeout(self, timeout):
            self.__timeout = timeout

        def assign(self, send_fd, recv_fd):
            with self.__cancel_lock:
                if self.__cancelled:
                    raise _make_cancelled_error()
                if not self.__pipe:
                    self.__create_pipe()
                self.__send_fd = send_fd
                self.__recv_fd = recv_fd

        def wait_recv(self):
            rlist, _, _ = select.select([self.__pipe[0], self.__recv_fd], [], [], self.__timeout)
            if self.__cancelled:
                raise _make_cancelled_error()
            if not rlist:
                raise _make_timeout_error()

        def wait_send(self):
            _, wlist, elist = select.select([self.__pipe[0]], [self.__send_fd], [self.__send_fd], self.__timeout)
            if self.__cancelled:
                raise _make_cancelled_error()
            if not wlist and not elist:
                raise _make_timeout_error()

        def cancel(self):
            with self.__cancel_lock:
                self.__cancelled = True
                if self.__pipe:
                    self.__pipe[1].send(b'\0')

        def is_cancelled(self):
            return self.__cancelled

        def close(self):
            with self.__cancel_lock:
                if self.__pipe:
                    self.__pipe[0].close()
                    self.__pipe[1].close()
                    self.__pipe = None
else:
    class _Poller:
        def __init__(self):
            self.__pipe = None
            self.__cancel_lock = threading.Lock()
            self.__cancelled = False
            self.__send_poll = None
            self.__recv_poll = None
            self.__timeout_ms = None

        def set_timeout(self, timeout):
            self.__timeout_ms = max(1, int(round(timeout * 1000)))

        def assign(self, send_fd, recv_fd):
            with self.__cancel_lock:
                if self.__cancelled:
                    raise _make_cancelled_error()
                if not self.__pipe:
                    self.__pipe = os.pipe()
                self.__send_poll = select.poll()
                self.__send_poll.register(self.__pipe[0], select.POLLIN)
                self.__send_poll.register(send_fd, select.POLLOUT)
                self.__recv_poll = select.poll()
                self.__recv_poll.register(self.__pipe[0], select.POLLIN)
                self.__recv_poll.register(recv_fd, select.POLLIN)

        def wait_recv(self):
            if not self.__recv_poll.poll(self.__timeout_ms):
                raise _make_timeout_error()
            if self.__cancelled:
                raise _make_cancelled_error()

        def wait_send(self):
            if not self.__send_poll.poll(self.__timeout_ms):
                raise _make_timeout_error()
            if self.__cancelled:
                raise _make_cancelled_error()

        def cancel(self):
            with self.__cancel_lock:
                self.__cancelled = True
                if self.__pipe:
                    os.write(self.__pipe[1], b'\0')

        def is_cancelled(self):
            return self.__cancelled

        def close(self):
            with self.__cancel_lock:
                if self.__pipe:
                    os.close(self.__pipe[0])
                    os.close(self.__pipe[1])
                    self.__pipe = None

        def __del__(self):
            self.close()


if _windows:
    # No support for timeouts and cancellation on Windows.
    class PipeChannel(Channel):
        def __init__(self, in_pipe, out_pipe, close, logger=_logger):
            self.__in_pipe = in_pipe
            self.__out_pipe = out_pipe
            self.__close = close

        def recv_into(self, buffer):
            return self.__in_pipe.readinto(buffer)

        def send(self, buffer):
            self.__out_pipe.write(buffer)
            self.__out_pipe.flush()
            return len(buffer)

        def close(self):
            if self.__close:
                self.__in_pipe.close()
                self.__out_pipe.close()
else:
    import fcntl

    class PipeChannel(Channel):
        def __init__(self, in_pipe, out_pipe, close, logger=_logger):
            out_pipe.flush()
            fcntl.fcntl(in_pipe, fcntl.F_SETFL, fcntl.fcntl(in_pipe, fcntl.F_GETFL) | os.O_NONBLOCK)
            self.__in_pipe = in_pipe
            self.__out_pipe = out_pipe
            self.__out_fd = out_pipe.fileno()
            self.__poller = _Poller()
            self.__poller.assign(self.__out_fd, self.__in_pipe.fileno())
            self.__close = close

        def set_timeout(self, timeout):
            self.__poller.set_timeout(timeout)

        def recv_into(self, buffer):
            nbytes = self.__in_pipe.readinto(buffer)
            if nbytes is not None:
                return nbytes
            self.__poller.wait_recv()
            nbytes = self.__in_pipe.readinto(buffer)
            assert nbytes is not None
            return nbytes

        def send(self, buffer):
            self.__poller.wait_send()
            return os.write(self.__out_fd, buffer)

        def cancel(self):
            self.__poller.cancel()

        def close(self):
            self.__poller.close()
            if self.__close:
                self.__in_pipe.close()
                self.__out_pipe.close()

        def __repr__(self):
            return '<{} {!r} {!r}>'.format(self.__class__.__name__, self.__in_pipe, self.__out_pipe)


class ProcessChannel(PipeChannel):
    def __init__(self, args=None, proc=None, logger=_logger, **popen_extras):
        if proc is None:
            if args is None:
                raise ValueError('Arguments must be provided if process is omitted')
            proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, **popen_extras)
            logger.debug('Started child process %r, pid=%r', args, proc.pid)
        else:
            if proc.stdin is None:
                raise ValueError('Process standart input is not a pipe')
            if proc.stdout is None:
                raise ValueError('Process standart output is not a pipe')
        super().__init__(proc.stdout, proc.stdin, close=True, logger=logger)
        self.__process = proc
        self.__args = args
        self.__logger = logger
        self.__timeout = None

    def set_timeout(self, timeout):
        self.__timeout = timeout
        super().set_timeout(timeout)

    def cancel(self):
        self.__kill_process()
        super().cancel()

    def __kill_process(self):
        if self.__process.poll() is not None:
            return
        self.__logger.debug('Killing process %r', self.__process.pid)
        try:
            self.__process.kill()
        except OSError:  # process already terminated
            pass

    def close(self):
        super().close()
        self.__logger.debug('Joining process %r', self.__process.pid)
        try:
            retcode = self.__process.wait(self.__timeout)
        except subprocess.TimeoutExpired:
            self.__kill_process()
            retcode = self.__process.wait()
        self.__logger.debug('Process joined, retcode=%r', retcode)
        if retcode:
            raise subprocess.CalledProcessError(retcode, self.__args)

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, self.__process.args)


_getaddrinfo = socket.getaddrinfo  # patched by tests


class SocketChannel(Channel):
    def __init__(self, family=None, address=None, sock=None, resolve=False, logger=_logger):
        if sock is None:
            if address is None:
                raise ValueError('Address must be provided if socket is omitted')
            if family is None:
                raise ValueError('Family must be provided if socket is omitted')
        else:
            if family is not None:
                raise ValueError('Family must be omitted if socket is provided')
            family = sock.family
        if address is None:
            if resolve:
                raise ValueError('Resolver may be used only when address is provided')
        self.__family = family
        self.__address = address
        self.__resolve = resolve
        self.__sock = sock
        self.__poller = _Poller()
        self.__logger = logger
        if self.__sock is not None:
            self.__prepare_socket()

    def set_timeout(self, timeout):
        self.__poller.set_timeout(timeout)

    def __prepare_socket(self):
        self.__tune_socket()
        self.__sock.setblocking(False)
        self.__poller.assign(self.__sock.fileno(), self.__sock.fileno())

    def __tune_socket(self):
        if self.__sock.family in (socket.AF_INET, socket.AF_INET6):
            self.__sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def __connect_once(self, family, address):
        if self.__sock is None:
            self.__sock = socket.socket(family, socket.SOCK_STREAM)
            self.__prepare_socket()
        try:
            self.__logger.debug('Trying to connect to %r', address)
            self.__sock.connect(address)
            return
        except BlockingIOError:
            pass
        self.__poller.wait_send()
        error = self.__sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if error != 0:
            raise OSError(error, os.strerror(error))

    def __connect_loop(self):
        if self.__resolve:
            host, port = self.__address
            addrinfo = _getaddrinfo(host, port, self.__family, socket.SOCK_STREAM)
        else:
            addrinfo = [(self.__family, None, None, None, self.__address)]

        last_error = None
        for family, _, _, _, address in addrinfo:
            had_socket = self.__sock is not None
            try:
                self.__connect_once(family, address)
                return
            except OSError as error:
                if self.__poller.is_cancelled():
                    raise
                if not had_socket and self.__sock is not None:
                    self.__sock.close()
                    self.__sock = None

                self.__logger.debug('connect throws %r', error)
                last_error = error

        if last_error is None:
            # C getaddrinfo() returns one or more items, so this should hopefully never happen
            raise RuntimeError('getaddrinfo returned empty list')
        else:
            try:
                raise last_error
            finally:
                last_error = None

    def connect(self):
        if self.__address:
            self.__connect_loop()

    def send(self, buffer):
        self.__poller.wait_send()
        return self.__sock.send(buffer)

    def recv_into(self, buffer):
        self.__poller.wait_recv()
        return self.__sock.recv_into(buffer)

    def cancel(self):
        self.__poller.cancel()

    def close(self):
        self.__poller.close()
        if self.__sock is not None:
            self.__sock.close()

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, self.__sock or self.__address)



import threading
import struct
import pickle
import sys
import types
import atexit
import io
import queue
import ast
import socket
import linecache
import os
import warnings
import copy
import itertools
import collections

from ._importer import get_bootloader, get_finder
from ._utils import (
    ignore_exception_at, Executor, Future, SimpleQueue, format_exception_only, logger as _logger,
    IdentityAdapter, atomic_count, set_debug_level)
from ._channels import PipeChannel, SocketChannel, ProcessChannel


if sys.version_info < (3, 4):
    warnings.warn('Python versions older than 3.4 are not supported')


_version = (0, 1, 0)
__version__ = '{}.{}.{}'.format(*_version)


class _Framer:
    def __init__(self, channel):
        self.__channel = channel

    def send_superblock(self, superblock):
        self.__sendall(superblock)

    def send_message(self, msg):
        buffers = msg.to_buffers()
        for num, buffer in enumerate(buffers, 1):
            self.__sendall(struct.pack('>Q', len(buffer) << 1 | (num < len(buffers))))
            self.__sendall(buffer)

    def send_bootloader(self, bootloader):
        data = bootloader.encode('ascii')
        self.__sendall(struct.pack('>I', len(data)) + data)

    def __sendall(self, buffer):
        buffer = memoryview(buffer)
        while buffer:
            size = self.__channel.send(buffer)
            assert size > 0
            buffer = buffer[size:]

    def recv_superblock(self, size):
        superblock = self.__recvall(size)
        if superblock is None:
            raise EOFError('Incomplete superblock')
        return superblock

    def recv_message(self):
        buffers = collections.deque()
        more = True
        while more:
            header = self.__recvall(8)
            if header is None:
                raise EOFError('Incomplete frame size')
            size, = struct.unpack('>Q', header)
            more = size & 1
            size >>= 1
            buffer = self.__recvall(size)
            if buffer is None:
                raise EOFError('Incomplete frame')
            buffers.append(buffer)
        return _Message(buffers)

    def __recvall(self, size):
        data = bytearray(size)
        buffer = memoryview(data)
        while buffer:
            size = self.__channel.recv_into(buffer)
            if size == 0:
                return None
            buffer = buffer[size:]
        return data


# Message consists of:
#   - an internal buffer that is read from / written to with get_*/put_* methods
#   - a list of out-of-band buffers accessible through 'buffers' attribute
class _Message:
    def __init__(self, buffers=None):
        if buffers:
            self.__buffer = io.BytesIO(buffers.popleft())
            self.buffers = buffers
        else:
            self.__buffer = io.BytesIO()
            self.buffers = collections.deque()

    def to_buffers(self):
        buffers = collections.deque(self.buffers)
        buffers.appendleft(self.__buffer.getbuffer())
        return buffers

    def put_uint(self, obj):
        if obj <= 0xfc:
            self.__buffer.write(struct.pack('>B', obj))
        elif obj <= 0xffff:
            self.__buffer.write(b'\xfd' + struct.pack('>H', obj))
        elif obj <= 0xffffffff:
            self.__buffer.write(b'\xfe' + struct.pack('>I', obj))
        else:
            self.__buffer.write(b'\xff' + struct.pack('>Q', obj))

    def get_uint(self):
        data = self.__buffer.read(1)
        if not data:
            return 0
        obj, = struct.unpack('>B', data)
        if obj <= 0xfc:
            return obj
        elif obj == 0xfd:
            return struct.unpack('>H', self.__buffer.read(2))[0]
        elif obj == 0xfe:
            return struct.unpack('>I', self.__buffer.read(4))[0]
        else:
            return struct.unpack('>Q', self.__buffer.read(8))[0]

    def put_bytes(self, obj):
        self.put_uint(len(obj))
        self.__buffer.write(obj)

    def get_bytes(self):
        return self.__buffer.read(self.get_uint())


class _ConnTls(threading.local):
    def __init__(self):
        self.conn = None
        self.proxy_ids = None


_conn_tls = _ConnTls()


class OperationError(Exception):
    pass


class DisconnectError(OperationError):
    pass


class ProtocolError(Exception):
    pass


_GREETING = 0
_APPLY = 1
_RESULT = 2
_GOODBYE = 3
_DEL_OBJS = 4
_IDLE = 5

_RESULT_SUCCESS = 2
_RESULT_ERROR = 3
_RESULT_OPERATION_ERROR = 4


class _ConnScope(object):
    def __init__(self, conn, proxy_ids=None):
        self.__conn = conn
        self.__proxy_ids = proxy_ids

    def __enter__(self):
        _conn_tls.conn, self.__conn = self.__conn, _conn_tls.conn
        _conn_tls.proxy_ids, self.__proxy_ids = self.__proxy_ids, _conn_tls.proxy_ids

    def __exit__(self, exc_type, exc_val, exc_tb):
        _conn_tls.conn = self.__conn
        _conn_tls.proxy_ids = self.__proxy_ids


_SenderQueue = SimpleQueue  # patched by tests


class _SenderThread(threading.Thread):
    def __init__(self, framer, cancel_func, logger):
        super(_SenderThread, self).__init__(daemon=True)
        self.__logger = logger
        self.__framer = framer
        self.__cancel_func = cancel_func
        self.__queue = _SenderQueue()
        self.__idle_timeout = None

    def run(self):
        try:
            while True:
                try:
                    task = self.__queue.get(self.__idle_timeout)
                except queue.Empty:
                    task = self.__idle
                if task() is False:
                    break
        # Catch exceptions raised by Channel.send(). Different channel types may use different types.
        except Exception as error:
            self.__cancel_func(error)

    def __idle(self):
        self.__logger.debug('Sending idle frame')
        msg = _Message()
        msg.put_uint(_IDLE)
        self.__framer.send_message(msg)

    def send_superblock(self, superblock):
        def handler():
            self.__framer.send_superblock(superblock)

        self.__queue.put(handler)

    # Important: this function must be reentrant
    def send_message(self, msg):
        def handler():
            self.__framer.send_message(msg)

        self.__queue.put(handler)

    def send_bootloader(self, bootloader):
        def handler():
            self.__framer.send_bootloader(bootloader)

        self.__queue.put(handler)

    def send_message_and_stop(self, msg):
        def handler():
            self.__framer.send_message(msg)
            return False

        self.__queue.put(handler)

    def stop(self):
        def handler():
            return False

        self.__queue.put(handler)

    def update_idle_timeout(self, timeout):
        def handler():
            self.__idle_timeout = timeout

        self.__queue.put(handler)


_SUPERBLOCK_SIZE = 16
_MAGIC = 0xe35b9e78
_BIOS_MAGIC = 0xe54e6d01


def _get_superblock():
    return struct.pack('>IBBB', _MAGIC, *_version) + b'\0' * (_SUPERBLOCK_SIZE - 7)


def _get_bios_superblock():
    return struct.pack('>IBBB', _BIOS_MAGIC, 0, 1, 0) + b'\0' * (_SUPERBLOCK_SIZE - 7)


def get_bios():
    to_skip = _SUPERBLOCK_SIZE + 4
    to_send = _get_bios_superblock()
    return ("import sys;i,o=sys.stdin.buffer,sys.stdout.buffer;o.write({!r});o.flush();i.read({!r});"
            "ACROSS='stdio',;exec(i.readline())".format(to_send, to_skip))


# timeout (in ms) is transmitted as uint64; these limits help avoiding internal
# errors when too small/large value is given; note that 1ms timeout will probably
# sooner or later cause a connection loss; we also have to ensure that system limits
# are not exceeded; e.g. threading.TIMEOUT_MAX is around 49 days on Windows
_DEFAULT_TIMEOUT = 60.0
_MIN_TIMEOUT = 0.001
_MAX_TIMEOUT = 86400.0  # one day


def _sanitize_timeout(timeout):
    if timeout is None:
        return timeout
    if isinstance(timeout, int):
        timeout = float(timeout)
    elif not isinstance(timeout, float):
        raise TypeError('timeout must be a float')
    if not timeout > 0.0:
        raise ValueError('timeout must be positive')
    return max(_MIN_TIMEOUT, min(_MAX_TIMEOUT, timeout))


class Options:
    _check_mode = False

    def __init__(self, **options):
        self.timeout = _DEFAULT_TIMEOUT
        self.accept_exported_modules = False

        self._assign(options)

    def copy(self, **options):
        other = copy.copy(self)
        other._assign(options)
        return other

    def _assign(self, options):
        for name, value in options.items():
            setattr(self, name, value)

    # Let's be nice enough to detect typos when setting options (e.g. 'timout')
    def __setattr__(self, name, value):
        if Options._check_mode and not hasattr(_default_options, name):
            raise AttributeError('No option named \'{}\''.format(name))
        super().__setattr__(name, value)

    def __repr__(self):
        options = []
        for option, value in sorted(self.__dict__.items()):
            if value != getattr(_default_options, option):
                options.append('{}={!r}'.format(option, value))
        return '{}({})'.format(self.__class__.__name__, ', '.join(options))


_default_options = Options()
Options._check_mode = True


def _sanitize_options(options=None):
    if options is None:
        return _default_options
    return options.copy(
        timeout=_sanitize_timeout(options.timeout),
    )


_unclosed_connections = set()
_connection_counter = atomic_count(1)

# Receiver thread establishes connection / performs handshake. Thre is no error (Connection.__error is None),
# and Connection.cancel() was not called. Sending data is allowed only from receiver thread.
_STARTING = 0
# Handshake completed. No error and no cancel. Any thread may send data.
_RUNNING = 1
# Receiver thread shuts down connection. Sender thread no longer accepts new data to send (new data is silently
# dropped). Some error might have occurred, or cancel() may have been called.
_STOPPING = 2
# Receiver thread finished execution and is waiting to be joined. All other resources are already freed.
# Some error might have occurred, or cancel() might have been called.
_STOPPED = 3
# Receiver thread was joined. Error, if any, has been re-raised to the user (from Connection.close()).
_CLOSED = 4


class Connection:
    def __init__(self, channel, *, options=None, on_stopped=None, logger=_logger):
        options = _sanitize_options(options)
        timeout = options.timeout
        if timeout is not None:
            channel.set_timeout(timeout)
            timeout_ms = max(1, int(round(timeout * 1000.0)))
        else:
            timeout_ms = 0

        if options.accept_exported_modules:
            finder = get_finder()
            finder.set_connection(self)
        else:
            finder = None

        self.__logger = logger
        self.__channel = channel
        self.__timeout_ms = timeout_ms
        self.__framer = _Framer(channel)
        self.__lock = threading.Lock()
        self.__sender = _SenderThread(self.__framer, logger=self.__logger, cancel_func=self.__cancel)
        self.__receiver_thread = threading.Thread(target=self.__receiver_loop, daemon=True)
        self.__state_condition = threading.Condition(self.__lock)
        self.__handlers_locked = self.__greeting_handlers_locked
        self.__state = _STARTING
        self.__was_running = False
        self.__cancelled = False
        self.__error = None
        self.__on_stopped = on_stopped
        self.__sender.send_superblock(_get_superblock())

        self.__calls = {}
        self.__next_call_id = 0
        self.__executor = Executor()

        self.__objs = {}
        self.__obj_counter = 0

        self._finder = finder

        _unclosed_connections.add(self)

        self.__logger.info('Connection starts, channel=%r, options=%r', channel, options)
        self.__receiver_thread.start()

    @classmethod
    def _make_logger(cls):
        return IdentityAdapter(_logger, 'conn/{}'.format(next(_connection_counter)))

    @classmethod
    def from_tcp(cls, host, port, **kwargs):
        logger = cls._make_logger()
        return cls(
            SocketChannel(family=socket.AF_UNSPEC, address=(host, port), resolve=True, logger=logger),
            logger=logger,
            **kwargs
        )

    @classmethod
    def from_unix(cls, path, **kwargs):
        if not hasattr(socket, 'AF_UNIX'):
            raise RuntimeError('Unix domain sockets are not available')
        logger = cls._make_logger()
        return cls(
            SocketChannel(family=socket.AF_UNIX, address=path, logger=logger),
            logger=logger,
            **kwargs
        )

    @classmethod
    def from_socket(cls, sock, **kwargs):
        logger = cls._make_logger()
        return cls(SocketChannel(sock=sock, logger=logger), logger=logger, **kwargs)

    @classmethod
    def from_pipes(cls, in_pipe, out_pipe, **kwargs):
        logger = cls._make_logger()
        return cls(PipeChannel(in_pipe, out_pipe, close=True, logger=logger), logger=logger, **kwargs)

    @classmethod
    def from_stdio(cls, **kwargs):
        logger = cls._make_logger()
        return cls(
            PipeChannel(sys.stdin.buffer, sys.stdout.buffer, close=False, logger=logger),
            logger=logger,
            **kwargs
        )

    @classmethod
    def from_command(cls, args, **kwargs):
        assert isinstance(args, (list, tuple))
        logger = cls._make_logger()
        return cls(ProcessChannel(args, logger=logger), logger=logger, **kwargs)

    @classmethod
    def from_shell(cls, script, **kwargs):
        assert isinstance(script, (str, bytes))
        logger = cls._make_logger()
        return cls(ProcessChannel(script, shell=True, logger=logger), logger=logger, **kwargs)

    @classmethod
    def from_process(cls, proc, **kwargs):
        logger = cls._make_logger()
        return cls(ProcessChannel(proc=proc, logger=logger), logger=logger, **kwargs)

    def _get_obj(self, id):
        return self.__objs[id]

    def _put_obj(self, obj):
        with self.__lock:
            id = self.__obj_counter
            self.__obj_counter += 1
        self.__objs[id] = obj
        return id

    # Important: this function must be reentrant
    def _del_remote_obj(self, id):
        msg = _Message()
        msg.put_uint(_DEL_OBJS)
        msg.put_uint(1)
        msg.put_uint(id)
        self.__sender.send_message(msg)

    def __enter__(self):
        with self.__lock:
            while self.__state == _STARTING or (self.__state == _STOPPING and not self.__was_running):
                self.__state_condition.wait()
            if self.__was_running and self.__state != _CLOSED:
                return self
            if self.__state != _RUNNING:
                self.__set_closed_locked()
                raise ValueError('Connection is closed')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def export(self, *modules):
        self.call(_export, modules)

    def close(self):
        with self.__lock:
            while self.__state == _STARTING:
                self.__state_condition.wait()
            self.__set_stopping_locked()
            while self.__state == _STOPPING:
                self.__state_condition.wait()
            self.__set_closed_locked()

    def __set_closed_locked(self):
        if self.__state != _CLOSED:
            self.__state = _CLOSED
            _unclosed_connections.remove(self)
            self.__receiver_thread.join()
            self.__logger.info('Connection closed')
            if self.__error is not None:
                try:
                    raise self.__error
                finally:
                    self.__error = None

    def cancel(self):
        self.__cancel(None)

    def __cancel(self, error):
        with self.__lock:
            if not self.__cancelled and self.__state not in (_STOPPED, _CLOSED):
                if error is None:
                    self.__logger.info('Cancelling connection')
                else:
                    self.__logger.error('Aborting connection due to %r', error)
                    self.__error = error
                self.__cancelled = True
                self.__set_stopping_locked()
                try:
                    self.__channel.cancel()
                # Channel.cancel() should rather not raise any exception, but we handle exceptions from other channel
                # methods, so let's not make cancel() any different.
                except Exception:
                    ignore_exception_at(self.__channel.cancel)

    def __set_stopping_locked(self):
        if self.__state in (_STARTING, _RUNNING):
            self.__state = _STOPPING
            self.__state_condition.notify_all()
            if not self.__cancelled:
                self.__logger.info('Stopping connection')
                msg = _Message()
                msg.put_uint(_GOODBYE)
                self.__sender.send_message_and_stop(msg)
            else:
                self.__sender.stop()
            for future in self.__calls.values():
                future.set_exception(self.__make_disconnect_error_locked())
            self.__calls.clear()

    def call(_self, _func, *args, **kwargs):
        future = Future()

        with _self.__lock:
            while _self.__state == _STARTING:
                _self.__state_condition.wait()
            if _self.__state != _RUNNING:
                raise _self.__make_disconnect_error_locked()
            call_id = _self.__next_call_id
            _self.__next_call_id += 1
            _self.__calls[call_id] = future

        try:
            msg = _Message()
            msg.put_uint(_APPLY)
            msg.put_uint(call_id)
            _self.__serialize(msg, (_func, args, kwargs))
        except:
            with _self.__lock:
                # May be already deleted if connection entered stopping state
                _self.__calls.pop(call_id, None)
            raise

        _self.__sender.send_message(msg)

        msg = future.result()
        result = msg.get_uint()
        if result == _RESULT_SUCCESS:
            return _self.__deserialize(msg)
        elif result == _RESULT_ERROR:
            raise _self.__deserialize(msg, as_exc=True)
        else:  # _RESULT_OPERATION_ERROR
            raise OperationError(msg.get_bytes().decode('utf-8', errors='replace'))

    def __make_disconnect_error_locked(self):
        if self.__state == _CLOSED:
            return DisconnectError('Connection is closed')
        elif not self.__cancelled:
            return DisconnectError('Connection was remotely closed')
        elif self.__error is not None:
            return DisconnectError('Connection was aborted due to {}'.format(
                format_exception_only(self.__error)))
        else:
            return DisconnectError('Connection was cancelled')

    def __serialize(self, msg, obj, as_exc=False):
        proxy_ids = []
        with _ConnScope(self, proxy_ids=proxy_ids):
            try:
                if as_exc:
                    obj = _dump_exception(obj)
                # Use pickle protocol added in Python 3.4.
                msg.buffers.append(pickle.dumps(obj, protocol=4))
                msg.put_uint(len(proxy_ids))
                for proxy_id in proxy_ids:
                    msg.put_uint(proxy_id)
            # Pickling may fail with all different exception types, let's catch them all.
            except Exception as error:
                for proxy_id in proxy_ids:
                    del self.__objs[proxy_id]
                raise OperationError('Pickling failed due to {}'.format(
                    format_exception_only(error))) from error

    def __deserialize(self, msg, as_exc=False):
        proxy_ids = []
        with _ConnScope(self, proxy_ids=proxy_ids):
            try:
                obj = pickle.loads(msg.buffers.popleft())
                if as_exc:
                    obj = _load_exception(obj)
                return obj
            # Unpickling may fail with all different exception types, let's catch them all.
            except Exception as error:
                leaked_proxy_ids = set(msg.get_uint() for _ in range(msg.get_uint())) - set(proxy_ids)
                if leaked_proxy_ids:
                    msg = _Message()
                    msg.put_uint(_DEL_OBJS)
                    msg.put_uint(len(leaked_proxy_ids))
                    for proxy_id in leaked_proxy_ids:
                        msg.put_uint(proxy_id)
                    self.__sender.send_message(msg)
                raise OperationError('Unpickling failed failed due to {}'.format(
                    format_exception_only(error))) from error

    def __receiver_loop(self):
        try:
            self.__channel.connect()
        # Different channel types may raise exceptions of different types.
        except Exception as error:
            self.__cancel(error)
        else:
            self.__sender.start()

            try:
                self.__receive_superblock()
                self.__receive_msgs()
            # Catch exceptions raised by Channel.recv() and ProtocolError. Different channel types use
            # different exception types.
            except Exception as error:
                self.__cancel(error)

            self.__sender.join()

        try:
            self.__channel.close()
        # Different channel types may raise exceptions of different types.
        except Exception as error:
            self.__cancel(error)

        self.__executor.close()

        with self.__lock:
            self.__logger.info('Connection stopped')
            self.__state = _STOPPED
            self.__state_condition.notify_all()

        if self.__on_stopped is not None:
            self.__on_stopped(self)

    def __receive_superblock(self):
        while True:
            superblock = self.__framer.recv_superblock(_SUPERBLOCK_SIZE)
            magic, major, minor, patch = struct.unpack_from('>IBBB', superblock)
            if magic == _MAGIC:
                self.__logger.debug('Remote across version is %s.%s.%s', major, minor, patch)
                if (major, minor) != _version[:2]:
                    raise ProtocolError('Local across {} is not compatible with remote {}.{}.{}'.format(
                        __version__, major, minor, patch
                    ))
                msg = _Message()
                msg.put_uint(_GREETING)
                msg.put_uint(self.__timeout_ms)
                self.__sender.send_message(msg)
                break
            elif magic == _BIOS_MAGIC:
                self.__logger.debug('Bootrapping needed with version at least %s.%s.%s', major, minor, patch)
                if (major, minor, patch) > _version:
                    raise ProtocolError('At least across {}.{}.{} is needed to bootstrap connection'.format(
                        major, minor, patch
                    ))
                options = Options(
                    timeout=self.__timeout_ms / 1000,
                    accept_exported_modules=True,
                )
                self.__sender.send_bootloader(get_bootloader(_start, options))
                self.__sender.send_superblock(_get_superblock())
            else:
                raise ProtocolError('Invalid magic: 0x{:x}'.format(magic))

    def __receive_msgs(self):
        while True:
            msg = self.__framer.recv_message()
            with self.__lock:
                if self.__cancelled:
                    break
                msg_type = msg.get_uint()
                handler = self.__handlers_locked.get(msg_type)
                if handler is None:
                    handlers_name = self.__handlers_locked[None]
                    raise ProtocolError('Invalid message in {} state: {}'.format(handlers_name, msg_type))
                if handler(self, msg) is False:
                    break

    def __handle_greeting_locked(self, msg):
        timeout_ms = msg.get_uint()
        if timeout_ms:
            # idle messages should be sent after a half of timeout passes
            idle_timeout = timeout_ms / 2000.0
        else:
            idle_timeout = None

        self.__logger.info('Connection is running, idle_timeout=%r', idle_timeout)
        self.__sender.update_idle_timeout(idle_timeout)
        self.__handlers_locked = self.__ready_handlers_locked
        self.__state = _RUNNING
        self.__was_running = True
        self.__state_condition.notify_all()

    def __handle_apply_locked(self, msg):
        if self.__state != _RUNNING:
            return
        self.__executor.submit(self.__process_apply, msg)

    def __process_apply(self, msg):
        call_id = msg.get_uint()
        try:
            # Deserialize
            obj = self.__deserialize(msg)

            # Execute
            with _ConnScope(self):
                try:
                    func, args, kwargs = obj
                    try:
                        value = func(*args, **kwargs)
                        result_type = _RESULT_SUCCESS
                    # If we allow raising OperationError between processes, then the other side wouldn't be able to
                    # determine whether the call failed to execute, or ended up generating OperationError. This
                    # is especially problematic for DisconnectError, which indicates that the connection is
                    # no longer usable.
                    except OperationError as error:
                        raise RuntimeError('Cannot raise {} between processes'.format(
                            error.__class__.__name__)) from error
                # Catch all exceptions raised by 'func' (this is e.g. how concurrent.futures.ThreadPoolExecutor works).
                except BaseException as error:
                    value = error
                    result_type = _RESULT_ERROR

            # Serialize
            msg = _Message()
            msg.put_uint(_RESULT)
            msg.put_uint(call_id)
            msg.put_uint(result_type)
            self.__serialize(msg, value, as_exc=result_type == _RESULT_ERROR)
        except OperationError as error:
            # Handle serialization errors. First, try to serialize the whole error. If that fails, give up using
            # pickle and send only the message.
            try:
                msg = _Message()
                msg.put_uint(_RESULT)
                msg.put_uint(call_id)
                msg.put_uint(_RESULT_ERROR)
                self.__serialize(msg, error, as_exc=True)
            except OperationError:
                msg = _Message()
                msg.put_uint(_RESULT)
                msg.put_uint(call_id)
                msg.put_uint(_RESULT_OPERATION_ERROR)
                msg.put_bytes(str(error).encode('utf-8', errors='replace'))

        self.__sender.send_message(msg)
        value = None  # break reference cycle then 'value' holds an exception

    def __handle_result_locked(self, msg):
        call_id = msg.get_uint()
        future = self.__calls.pop(call_id, None)
        if future is None:
            raise ProtocolError('Call not found: {}'.format(call_id))
        future.set_result(msg)

    def __handle_del_objs_locked(self, msg):
        for _ in range(msg.get_uint()):
            del self.__objs[msg.get_uint()]

    def __handle_goodbye_locked(self, msg):
        self.__set_stopping_locked()
        return False

    def __handle_idle_locked(self, msg):
        self.__logger.debug('Received idle frame')

    __ready_handlers_locked = {
        _APPLY: __handle_apply_locked,
        _RESULT: __handle_result_locked,
        _DEL_OBJS: __handle_del_objs_locked,
        _GOODBYE: __handle_goodbye_locked,
        _IDLE: __handle_idle_locked,
        None: 'ready',
    }

    __greeting_handlers_locked = {
        _GREETING: __handle_greeting_locked,
        None: 'greeting',
    }

    def is_connected(self):
        # Lock not needed, because attribute access is atomic
        return self.__state in (_STARTING, _RUNNING)

    def wait(self):
        with self.__lock:
            while self.__state in (_STARTING, _RUNNING):
                self.__state_condition.wait()

    def call_ref(_self, _func, *args, **kwargs):
        return _self.call(_apply_ref, _func, args, kwargs)

    def replicate(self, obj):
        return self.call(ref, obj)

    def execute(_self, _source, **kwargs):
        return _self.call(_execute, _source, kwargs)


def _export(modules):
    finder = get_connection()._finder
    if finder is None:
        raise ValueError('Exporting modules is not allowed')
    finder.export(modules)


def _shutdown():
    for conn in _unclosed_connections.copy():
        conn.cancel()
        try:
            conn.close()
        # Connection.close() will re-raise i.e. any exceptions generated by Channel objects (which use various
        # exception types).
        except Exception:
            ignore_exception_at(conn.close)


atexit.register(_shutdown)


# Start a remotely bootstrapped connection. 'args' is the value of 'ACROSS' variable set by BIOS scripts.
def _start(options, args):
    _logger.debug('Creating bootstrapped connection, options=%s, args=%s', options, args)
    if args[0] == 'stdio':
        conn = Connection.from_stdio(options=options)
    elif args[0] == 'socket':
        conn = Connection.from_socket(args[1], options=options)
    else:
        raise RuntimeError('Unrecognized channel type {!r}, args={!r}'.format(args[0], args))
    with conn:
        conn.wait()


def _execute(source, scope):
    filename = '<execute>'
    varname = '__across_ret'
    mod = compile(source, filename, 'exec', ast.PyCF_ONLY_AST)
    if mod.body and isinstance(mod.body[-1], ast.Expr):
        mod.body[-1] = ast.Assign(
            [ast.Name(varname, ast.Store(), lineno=0, col_offset=0)],
            mod.body[-1].value, lineno=0, col_offset=0)
    exec(compile(mod, filename, 'exec'), scope)
    return scope.get(varname)


def _load_exception(obj):
    if not isinstance(obj, tuple):  # None or bare exception
        return obj
    exc = obj[0]
    exc.__context__ = _load_exception(obj[1])
    exc.__cause__ = _load_exception(obj[3])
    exc.__suppress_context__ = obj[2]
    exc.__traceback__ = _load_traceback(obj[4])
    return exc


def _dump_exception(exc, memo=None):
    if exc is None:
        return None
    if memo is None:
        memo = set()
    # It can happen that __context__ and __cause__ are the same, better not dump same exception multiple times.
    if id(exc) in memo:
        return exc
    memo.add(id(exc))
    return (
        exc,
        _dump_exception(exc.__context__, memo),
        exc.__suppress_context__,
        _dump_exception(exc.__cause__, memo),
        _dump_traceback(exc.__traceback__),
    )


def _dump_traceback(tb):
    entries = []
    base_packed_tb = None
    while tb is not None:
        frame = tb.tb_frame
        globs = frame.f_globals
        if tb.tb_next is None and _packed_tb_var in globs:
            base_packed_tb = globs[_packed_tb_var]
        else:
            code = frame.f_code
            filename, lineno = code.co_filename, tb.tb_lineno
            line = linecache.getline(filename, lineno, globs)
            entries.append((filename, lineno, code.co_name, line))
        tb = tb.tb_next
    packed_tb = []
    if entries:
        packed_tb.append((socket.gethostname(), os.getpid(), entries))
    if base_packed_tb:
        packed_tb.extend(base_packed_tb)
    return packed_tb


_packed_tb_var = '__across_packed_tb'
_code_tpl = compile('raise ValueError', '<traceback loader>', 'exec')


# This function creates a types.TracebackType object (which requires raising some exception, by the way)
# that is specifically crafted so that when 'traceback' module tries to format it, 'formatted_tb' string
# is printed as well.
def _generate_traceback_object(formatted_tb, packed_tb):
    co_name = '{}\n{}'.format(_code_tpl.co_name, formatted_tb.rstrip())
    try:
        replace = _code_tpl.replace  # since Python 3.8
    except AttributeError:
        # argument order is described in help(types.CodeType)
        code = types.CodeType(
            _code_tpl.co_argcount,
            _code_tpl.co_kwonlyargcount,
            _code_tpl.co_nlocals,
            _code_tpl.co_stacksize,
            _code_tpl.co_flags,
            _code_tpl.co_code,
            _code_tpl.co_consts,
            _code_tpl.co_names,
            _code_tpl.co_varnames,
            _code_tpl.co_filename,
            co_name,
            _code_tpl.co_firstlineno,
            _code_tpl.co_lnotab,
            _code_tpl.co_freevars,
            _code_tpl.co_cellvars,
        )
    else:
        code = replace(co_name=co_name)

    try:
        exec(code, {_packed_tb_var: packed_tb})
    except ValueError:
        return sys.exc_info()[2].tb_next


def _load_traceback(packed_tb):
    buf = []
    max_repeats = 3  # same as in traceback module
    for hostname, pid, entries in packed_tb:
        buf.append('  [Below lines obtained from host "{}", process {}]\n'.format(hostname, pid))
        for (filename, lineno, name, line), group in itertools.groupby(entries):
            overflow = -max_repeats
            for _ in group:
                overflow += 1
                if overflow <= 0:
                    buf.append('  File "{}", line {}, in {}\n'.format(filename, lineno, name))
                    if line:
                        buf.append('    {}\n'.format(line.strip()))
            if overflow > 0:
                buf.append('  [Previous line repeated {} more {}]\n'.format(
                    overflow,
                    'time' if overflow == 1 else 'times',
                ))
    return _generate_traceback_object(''.join(buf), packed_tb)


class Proxy(object):
    def __init__(self, proxy_id):
        self._Proxy__conn = get_connection()
        self.__id = proxy_id

    def __deepcopy__(self, memo):
        return self._Proxy__conn.call(_get_obj, self.__id)

    def __reduce__(self):
        if _conn_tls.proxy_ids is None:
            raise RuntimeError("Only 'across.Connection' objects can pickle 'across.Proxy' objects")
        conn = _conn_tls.conn
        if conn is not self._Proxy__conn:
            raise RuntimeError("Proxy {!r} can only be pickled by {!r} connection, not {!r}".format(
                self, self._Proxy__conn, conn))
        return _get_obj, (self.__id, )

    def __del__(self):
        self._Proxy__conn._del_remote_obj(self.__id)


def get_proxy_connection(proxy):
    if not isinstance(proxy, Proxy):
        raise TypeError('Not a proxy: {!r}'.format(proxy))
    return proxy._Proxy__conn


_proxy_types = {}
_safe_magic = frozenset([
    '__contains__', '__delitem__', '__getitem__', '__len__', '__setitem__', '__call__', '__iter__', '__next__'])


def _get_methods(cls):
    return [
        attr
        for attr in dir(cls)
        if callable(getattr(cls, attr)) and (not attr.startswith('_') or attr in _safe_magic)
    ]


def _make_auto_proxy_type(methods):
    scope = {'Proxy': Proxy, '_apply_method': _apply_method, '_SELF': _SELF}
    source = ['class _AutoProxy(Proxy):\n']

    for meth in methods:
        source.append(
            '    def {0}(self, *args, **kwargs):\n'
            '        result = self._Proxy__conn.call(_apply_method, self, {0!r}, args, kwargs)\n'
            '        return self if result is _SELF else result\n'.format(meth)
        )

    if len(source) == 1:
        source.append('    pass\n')

    exec(''.join(source), scope)
    return scope['_AutoProxy']


class Reference(object):
    def __init__(self, obj):
        assert not isinstance(obj, (Proxy, Reference))
        self.__obj = obj

    def __reduce__(self):
        if _conn_tls.proxy_ids is None:
            raise RuntimeError("Only 'across.Connection' objects can pickle 'across.Reference' objects")
        methods = _get_methods(type(self.__obj))
        proxy_id = _conn_tls.conn._put_obj(self.__obj)
        _conn_tls.proxy_ids.append(proxy_id)
        return _make_proxy, (proxy_id, methods)


def ref(obj):
    return Reference(obj)


class _SELF:
    pass


def _apply_method(obj, name, args, kwargs):
    result = getattr(obj, name)(*args, **kwargs)
    if result is obj:
        return _SELF
    elif name == '__iter__':
        return ref(result)
    else:
        return result


def _get_obj(id):
    return get_connection()._get_obj(id)


def _apply_ref(func, args, kwargs):
    return ref(func(*args, **kwargs))


def _make_proxy(proxy_id, methods):
    methods = frozenset(methods)
    proxy_type = _proxy_types.get(methods)
    if proxy_type is None:
        proxy_type = _proxy_types[methods] = _make_auto_proxy_type(methods)
    _conn_tls.proxy_ids.append(proxy_id)
    return proxy_type(proxy_id)


def get_connection():
    conn = _conn_tls.conn
    if conn is None:
        raise ValueError('Current thread is not associated with any connection')
    return conn


# Export
set_debug_level = set_debug_level
